<p class="help-block" translate>
    The RADIUS token forwards the authentication request to another
    RADIUS server. You can choose if the PIN should be stripped and
    checked locally.
</p>

<h4 translate>Token data</h4>

<div ng-hide="form['radius.server']" translate class="bg-info">
    To ease the enrolling of RADIUS tokens you can
    set default values in the
    <a ui-sref="config.tokens({tokentype:'radius'})">RADIUS Token Config</a>.
</div>

<div class="form-group">
    <input type="checkbox" ng-model="form['radius.local_checkpin']"
           name="localPin" id="localPin">
    <label for="localPin" translate>Check the PIN locally</label>
</div>

<div class="form-group">
    <label for="radiusServer" translate>RADIUS Server</label>
    <input type="text"
           required
           class="form-control"
           placeholder="your.radius.server:1812"
           ng-model="form['radius.server']" name="radiusServer">
    <p class="help-block" translate>
        The RADIUS server may include the port number.
    </p>
</div>

<div class="form-group">
    <label for="radiusUser" translate>RADIUS User</label>
    <input type="text" ng-model="form['radius.user']"
           class="form-control"
           required
           plasceholder="{{ 'The username on the RADIUS server'|translate }}"
           name="radiusUser">
</div>

<div class="form-group">
    <label for="radiusSecret" translate>RADIUS Secret</label>
    <input type="text" ng-model="form['radius.secret']"
           class="form-control"
           required
           plasceholder="testing123"
           name="radiusSecret">
</div>
